import time
from ..models import vectorise
from .documents import Document, ElasticDocument
from elasticsearch import Elasticsearch, helpers
from typing import Dict, List
import shortuuid
from ..utils import elastic_to_search_results, check_duplicate_documents, \
    check_document_types, batch_items


class SearchResult:
    """Basic Kiri search result

    Attributes:
        document: Document object associated with result
        score: Float representing relevancy to search query
        preview: Text content "peek" at result (metatext)
    """

    def __init__(self, document: Document, score: float, preview: str = None):
        self.document = document
        self.score = score
        self.preview = preview

    def to_json(self, exclude_vectors=True):
        """Gets JSON form of search result

        Attributes:
            exclude_vectors: Exclude content and chunk vectors from the json output

        Returns:
            __dict__ attribute of SearchResult object
        """
        json_repr = vars(self)
        json_repr["document"] = json_repr["document"].to_json(
            exclude_vectors=exclude_vectors)

        return json_repr


class SearchResults:
    """List of SearchResults with search metadata

    Attributes:
        max_score: Highest score returned by the search
        total_results: Count of documents found
        results: List of <total_results> SearchResult objects
    """

    def __init__(self, max_score: float, total_results: int, results: List[SearchResult]):
        self.max_score = max_score
        self.total_results = total_results
        self.results = results

    def to_json(self, exclude_vectors=True):
        """Gets JSON form of returned results

        Attributes:
            exclude_vectors: Exclude content and chunk vectors from the json output

        Returns:
            __dict__ attribute of SearchResults object

        """
        json_repr = vars(self)
        json_repr["results"] = [
            r.to_json(exclude_vectors=exclude_vectors) for r in json_repr["results"]]
        return json_repr


class DocStore:
    """Base DocStore class for extension.

    Raises:
        NotImplementedError: If core init, upload, or search functions are not implemented
    """

    def __init__(self, *args, **kwargs):
        raise NotImplementedError("__init__ is not implemented!")

    def upload(self, *args, **kwargs):
        raise NotImplementedError("upload is not implemented!")

    def search(self, *args, **kwargs):
        raise NotImplementedError("search is not implemented!")


class InMemoryDocStore(DocStore):
    """DocStore variant for in memory document storage.
    Meant to be used for development and testing.
    """

    def __init__(self, doc_class: Document = Document):
        self.documents = []
        self._doc_class = doc_class

    def upload(self, documents: List[Document], vectorise_func,
               vectorise_model) -> None:

        # Check ID uniqueness
        check_duplicate_documents(documents)
        # Check type consistency
        check_document_types(documents)
        # Batching
        batches = batch_items(documents)

        for batch in batches:
            vectorise_func(batch, vectorise_model)
            self.documents += batch

    def search(self, query, vectorise_model, max_results=10, min_score=0.0,
               ids=None, body=None):
        query_vec = vectorise(query)

        documents = self.documents

        # Filter by id
        if ids:
            documents = [d for d in documents if d.id in ids]

        results = [SearchResult(d, -1.0) for d in documents]
        search_results = SearchResults(-1.0, len(results), results)
        return search_results, query_vec


class ElasticDocStore(DocStore):
    """DocStore variant for an Elasticsearch backend.

    Attributes:
        url: Location at which Elasticsearch is running
        index: Elastic index (database) to be used for this doc store
        doc_class: Document object type being stored in this index
    """

    def __init__(self, url: str, index: str = "kiri_default",
                 doc_class: ElasticDocument = ElasticDocument):
        self._client = Elasticsearch([url])
        self._index = index
        self._doc_class = doc_class

        correct_mapping = doc_class.elastic_mappings()

        # Setup index with correct mappings
        if self._client.indices.exists(self._index):
            # Check if mapping is correct
            mapping = self._client.indices.get_mapping(
                self._index).get(self._index).get("mappings")

            # Try updating mapping
            if mapping != correct_mapping:
                self._client.indices.close(self._index)
                self._client.indices.put_mapping(
                    correct_mapping, index=self._index)
                self._client.indices.open(self._index)

        else:
            # Create index with mapping
            self._client.indices.create(
                self._index, body={"mappings": correct_mapping})

    def upload(self, documents: List[ElasticDocument], vectorise_func, vectorise_model, index: str = None) -> None:
        """Upload documents to Elasticsearch

        Args:
            documents: List of documents to be uploaded to backend
            vectorise_func: Function used to vectorise document contents
            vectorise_model: Sentence transformer model name used for vectorisation
            index: Index (db) to be used -- uses initialized default if none provided
        """
        if not index:
            index = self._index

        # Check ID uniqueness
        check_duplicate_documents(documents)
        # Check type consistency
        check_document_types(documents)
        # Batching
        batches = batch_items(documents)

        for batch in batches:
            payload = []
            # Calculate vectors
            vectorise_func(batch, vectorise_model)

            for document in batch:
                # JSON representation of document
                doc_json = document.to_elastic()

                # Add correct index
                doc_json["_index"] = index

                # Rename id key
                doc_json["_id"] = doc_json["id"]
                del doc_json["id"]

                payload.append(doc_json)

                # Bulk upload to elasticsearch
                helpers.bulk(self._client, payload)

        # Update index
        self._client.indices.refresh(index=self._index)

    def search(self, query, vectorise_model, max_results=10, min_score=0.0, ids: List[str] = [], body=None):
        """Search documents from Elasticsearch

        Args:
            query: Question from which search is based
            vectorise_model: NLP model used to vectorise the query -- should match one used on docs
            max_results: Maximum number of search results to return
            min_score: Minimum relevancy score required to be included in results
            ids: Filter search results by ids
            body: Elasticsearch parameters to be used in search -- default made if none provided

        Returns:
            Tuple with the SearchResults object and the vectorised query
        """
        query_vec = vectorise(query)

        # elasticsearch does not support negative scores
        score_modifier = 1.0

        if body is None:
            q = {
                "bool": {
                    "should": [
                        {"match": {
                            "content": query
                        }},
                        {"match_all": {}}

                    ]
                }
            }

            if ids:
                q = {
                    "bool": {
                        "must": [
                            {
                                "terms": {
                                    "_id": ids
                                }
                            },
                            q
                        ]
                    }
                }

            body = {
                "min_score": min_score + score_modifier,
                "size": max_results,
                "query": {
                    "function_score": {
                        "query": q,
                        "script_score": {
                            "script": {
                                "source": f"(cosineSimilarity(params.query_vector, 'vector') + {score_modifier}) * (_score + 1)",
                                "params": {
                                    "query_vector": query_vec.tolist()}
                            }
                        },
                    }
                },
                "highlight": {
                    "pre_tags": ["<b>"],
                    "post_tags": ["</b>"],
                    "fragment_size": 100,
                    "fields": {
                        "content": {}
                    }
                },

            }

        start = time.time()
        res = self._client.search(index=self._index, body=body)
        search_results = elastic_to_search_results(
            res, score_modifier, self._doc_class)

        return search_results, query_vec



from typing import Callable, Dict, List, Tuple
from sentence_transformers import SentenceTransformer
from transformers import T5ForConditionalGeneration, T5Tokenizer

from .search import DocStore, SearchResults, Document, InMemoryDocStore
from .utils import process_documents, process_results
from .models import qa, summarise, emotion, zero_shot


class Kiri:
    """Core class of natural language engine

    Attributes:
        store: DocStore object to be used as the engine backend
        vectorise_model: Name of the SentenceTransformer model to be used in operations
        qa_model: Name of HuggingFace model to be used for Question/Answer
        process_doc_func: Function to be used when vectorising uploaded documents
        process_results_func: Function to be used for calculating final scores of results

    """

    def __init__(self, store: DocStore = None, vectorise_model: str = None,
                 qa_model: str = None,
                 process_doc_func: Callable[[
                     Document, str], List[float]] = None,
                 process_results_func: Callable[[SearchResults, str], None] = None):

        if store is None:
            store = InMemoryDocStore()

        if process_doc_func is None:
            # Use default vectoriser
            process_doc_func = process_documents

        if process_results_func is None:
            process_results_func = process_results

        self._store = store
        self._process_doc_func = process_doc_func
        self._process_results_func = process_results
        self._vectorise_model = vectorise_model

    def upload(self, documents: List[Document]) -> None:
        """Uploads documents to store

        Args:
            documents: List of documents for upload

        """

        return self._store.upload(documents, self._process_doc_func,
                                  self._vectorise_model)

    def search(self, query: str, max_results=10, min_score=0.0,
               preview_length=100, ids: List[str] = [], body=None) -> SearchResults:
        """Search documents from document store

        Args:
            query: Search string on which search is performed
            max_results: Maximum amount of documents to be returned from search
            min_score: Minimum score required to be included in results
            preview_length: Number of characters in the preview/metatext of results
            ids: 
            body: Elasticsearch request body to be passed to the backend

        """
        search_results, query_vec = self._store.search(query, self._vectorise_model,
                                                       max_results=max_results, min_score=min_score,
                                                       ids=ids, body=body)
        self._process_results_func(
            search_results, query_vec, self._store._doc_class,
            preview_length, max_results, min_score)
        return search_results

    def qa(self, question: str, context: str = None,
           prev_qa: List[Tuple[str, str]] = [], context_doc: Document = None):
        """Perform QA, either on docstore or on provided context.

        """
        if context_doc or context:
            c_string = context if context else context_doc.content
            return qa(question, c_string, prev_qa=prev_qa)
        else:
            search_results = self.search(question)
            answers = []
            for result in search_results.results[:3]:
                c_string = result.document.content
                answer = qa(question, c_string, prev_qa=prev_qa)
                answers.append(answer)
            return list(zip(answers, search_results.results[:3]))

    def summarise(self, input_text):
        # if type(input_text) != str:
        #     raise TypeError("input_text must be a string")

        # if input_text == "":
        #     raise ValueError("input_text must not be an empty string")

        return summarise(input_text)

    def emotion(self, input_text):
        # if type(input_text) != str:
        #     raise TypeError("input_text must be a string")

        # if input_text == "":
        #     raise ValueError("input_text must not be an empty string")

        return emotion(input_text)

    def classify(self, input_text, labels: List[str]):
        # if type(input_text) != str:
        #     raise TypeError("input_text must be a string")

        # if input_text == "":
        #     raise ValueError("input_text must not be an empty string")

        # if not isinstance(labels, list):
        #     raise TypeError("labels must be a list of strings")

        # if len(labels) == 0:
        #     raise ValueError("labels must contain at least one label")

        return zero_shot(input_text, labels)



from typing import List
from ..search import Document


def check_duplicate_documents(documents: List[Document]):
    seen = set()
    duplicates = set()
    for d in documents:
        if d.id not in seen:
            seen.add(d.id)
        else:
            duplicates.add(d.id)

    duplicates = list(duplicates)

    if len(duplicates) != 0:
        raise ValueError(
            f"list of documents contains duplicate ids, example: {duplicates[0]}")


def check_document_types(documents: List[Document]):
    # set
    types = {type(d) for d in documents}
    if len(types) != 1:
        raise ValueError("all documents must have the same type")



from typing import List, Tuple
from .generation import generate


def process_item(item):
    return f"summarise: {item}"


def summarise(input_text,
              model_name: str = None, tokenizer_name: str = None,
              local: bool = True):
    if local:
        if isinstance(input_text, list):
            # Process according to the model used
            input_text = [process_item(item) for item in input_text]
        else:
            input_text = process_item(input_text)

        return generate(input_text, model_name=model_name,
                        tokenizer_name=tokenizer_name, local=local)
    else:
        raise ValueError("Non local inference is not implemented!")



from typing import Dict


def elastic_to_search_results(res: Dict, score_modifier: float, doc_class):
    """Converts Elasticsearch result to Kiri SearchResults object

    Args:
        res: dictionary/JSON returned from Elasticsearch
        score_modifier: value to be subtracted from each result's score
        doc_class: Type of Document object to be included in SearchResult object

    Returns:
        SearchResults equivalent of Elastic result
    """

    from ..search import SearchResult, SearchResults

    hits = res.get("hits")

    max_score = hits.get("max_score")
    total_results = hits.get("total").get("value")

    search_results = []
    for hit in hits.get("hits"):
        source = hit["_source"]

        # score_modifier was added when searching on elasticsearch
        score = hit["_score"] - score_modifier

        document = doc_class.from_elastic(id=hit.get("_id"),
                                          **source)
        search_result = SearchResult(document, score)
        search_results.append(search_result)

    return SearchResults(max_score, total_results, search_results)



from typing import List, Tuple

DEFAULT_MODEL = "facebook/bart-large-mnli"
DEFAULT_TOKENIZER = "facebook/bart-large-mnli"
model = None
tokenizer = None


def calculate_probability(input_text, label):
    hypothesis = f"This example is {label}."
    features = tokenizer.encode(input_text, hypothesis, return_tensors="pt",
                                truncation_strategy="only_first")
    logits = model(features)[0]
    entail_contradiction_logits = logits[:, [0, 2]]
    probs = entail_contradiction_logits.softmax(dim=1)
    prob_label_is_true = probs[:, 1]
    return prob_label_is_true.item()


def zero_shot(input_text, labels: List[str], model_name: str = None,
              tokenizer_name: str = None, local: bool = True):
    # Refer to global variables
    global model
    global tokenizer
    # Setup
    if local:
        # Initialise model
        if model == None:
            from transformers import AutoModelForSequenceClassification
            # Use the default model
            if model_name == None:
                model = AutoModelForSequenceClassification.from_pretrained(
                    DEFAULT_MODEL)
            # Use the user defined model
            else:
                model = AutoModelForSequenceClassification.from_pretrained(
                    model_name)

        # Initialise tokenizer
        if tokenizer == None:
            from transformers import AutoTokenizer
            # Use the default tokenizer
            if tokenizer_name == None:
                tokenizer = AutoTokenizer.from_pretrained(DEFAULT_TOKENIZER)
            # Use the user defined tokenizer
            else:
                tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)

        if isinstance(input_text, list):
            # Must have a consistent amount of examples
            assert(len(input_text) == len(labels))
            # TODO: implement proper batching
            results_list = []
            for text, labels in zip(input_text, labels):
                results = {}
                for label in labels:
                    results[label] = calculate_probability(text, label)

                results_list.append(results)

            return results_list
        else:
            results = {}
            for label in labels:
                results[label] = calculate_probability(input_text, label)

            return results

    else:
        raise ValueError("Non local inference is not implemented!")



from typing import List, Tuple

DEFAULT_MODEL = "msmarco-distilroberta-base-v2"
model = None


def vectorise(input_text, model_name: str = None,
              local: bool = True):
    # Refer to global variables
    global model
    # Setup
    if local:
        # Initialise model
        if model == None:
            from sentence_transformers import SentenceTransformer
            # Use the default model
            if model_name == None:
                model = SentenceTransformer(
                    DEFAULT_MODEL)
            # Use the user defined model
            else:
                model = SentenceTransformer(model_name)

        return model.encode(input_text)

    else:
        raise ValueError("Non local inference is not implemented!")



from typing import List, Tuple

DEFAULT_MODEL = "kiri-ai/t5-base-qa-summary-emotion"
DEFAULT_TOKENIZER = "t5-base"
model = None
tokenizer = None


def generate(input_text, model_name: str = None, tokenizer_name: str = None,
             local: bool = True):
    # Refer to global variables
    global model
    global tokenizer
    # Setup
    if local:
        # Initialise model
        if model == None:
            from transformers import T5ForConditionalGeneration
            # Use the default model
            if model_name == None:
                model = T5ForConditionalGeneration.from_pretrained(
                    DEFAULT_MODEL)
            # Use the user defined model
            else:
                model = T5ForConditionalGeneration.from_pretrained(model_name)

        # Initialise tokenizer
        if tokenizer == None:
            from transformers import T5Tokenizer
            # Use the default tokenizer
            if tokenizer_name == None:
                tokenizer = T5Tokenizer.from_pretrained(DEFAULT_TOKENIZER)
            # Use the user defined tokenizer
            else:
                tokenizer = T5Tokenizer.from_pretrained(tokenizer_name)

        is_list = False
        if isinstance(input_text, list):
            is_list = True

        features = tokenizer(input_text, padding=True, return_tensors='pt')
        tokens = model.generate(input_ids=features['input_ids'],
                                attention_mask=features['attention_mask'], max_length=128)
        if is_list:
            return [tokenizer.decode(tokens, skip_special_tokens=True) for tokens in tokens]
        else:
            return tokenizer.decode(tokens[0], skip_special_tokens=True)

    else:
        raise ValueError("Non local inference is not implemented!")



from typing import List, Tuple
from .generation import generate


def process_item(question, context, prev_qa):
    input_text = [f"q: {qa[0]} a: {qa[1]}" for qa in prev_qa]
    input_text.append(f"q: {question}")
    input_text.append(f"c: {context}")
    input_text = " ".join(input_text)

    return input_text


def qa(question, context, prev_qa: List[Tuple[str, str]] = [],
       model_name: str = None, tokenizer_name: str = None,
       local: bool = True):
    if local:
        if isinstance(question, list):
            # Must have a consistent amount of examples
            assert(len(question) == len(context))
            if len(prev_qa) != 0:
                assert(len(question) == len(prev_qa))
            else:
                prev_qa = [prev_qa] * len(question)

            # Process according to the model used
            input_text = [process_item(q, c, p)
                          for q, c, p in zip(question, context, prev_qa)]
        else:
            input_text = process_item(question, context, prev_qa)

        return generate(input_text, model_name=model_name,
                        tokenizer_name=tokenizer_name, local=local)
    else:
        raise ValueError("Non local inference is not implemented!")



from sentence_transformers import SentenceTransformer
from nltk import sent_tokenize
from scipy.spatial.distance import cdist
from typing import List, Tuple, Union

from ..search import Document, ChunkedDocument
from ..models import vectorise


def batch_items(items, batch_size=16):
    batches = [items[i:i + batch_size]
               for i in range(0, len(items), batch_size)]
    return batches


def get_sentences(content):
    """Tokenizes document content into sentences

    Args:
        content: String of document contents

    Returns:
        list of strings, the separate sentences of <content>
    """
    sentences = sent_tokenize(content)

    return sentences


def chunk_document(document: ChunkedDocument):
    """Gets document chunks based on chunking level

    Args:
        document: Document to be chunked 

    Returns:
        Chunks to be assigned to <document>
    """

    chunking_level = document.chunking_level

    # Split to sentences
    sentences = get_sentences(document.content)

    # Join groups of sentences into chunks by chunking level
    # ["a", "b", "c"] with chunking level 2
    # creates ["a b", "c"]
    chunks = [" ".join(sentences[i:i + chunking_level])
              for i in range(0, len(sentences), chunking_level)]

    return chunks


def process_documents(documents: Union[Document, List[Document]], model_name: str):
    """Processes document based on type

    Args:
        documents: Document or list of Documents to be processed
        model: SentenceTransformer model name used for processing

    Raises:
        ValueError: If the given document's type doesn't have a vectorisation function
    """
    # Example for instance verification
    doc_example = documents[0]
    if isinstance(doc_example, Document):
        contents = [d.content for d in documents]
        content_vectors = vectorise(contents, model_name=model_name)

        # Set vectors for each document
        for document, vector in zip(documents, content_vectors):
            document.vector = vector

    # TODO: Batching for document chunks
    if isinstance(doc_example, ChunkedDocument):
        for document in documents:
            chunks = chunk_document(document)
            chunk_vectors = vectorise(chunks, model_name=model_name)
            document.chunks = chunks
            document.chunk_vectors = chunk_vectors

    if not isinstance(doc_example, Document) and not isinstance(doc_example, ChunkedDocument):
        raise ValueError(
            f"vectorisation of document of type {type(document)} is not implemented")


def calc_doc_score(max_score: float, doc_score: float, chunk_scores: Tuple[str, float], top_chunks: int = 3):
    # def calc_doc_score(doc_score: float, chunk_scores: Tuple[str, float], top_chunks: int = 3):
    """Calculates the overall score of a chunked document

    Args:
        max_score: The maximum score returned by the search, for normalization
        doc_score: The base relevancy score from Elasticsearch
        chunk_scores: Tuple of each chunk with its corresponding score
        top_chunks: number of chunks to be included in the weighting

    Returns:
        The final weighted score for the document
    """

    doc_scores = []

    norm_doc_score = doc_score / max_score

    for chunk, score in chunk_scores[:top_chunks]:
        doc_scores.append(score)

    avg_chunk_score = sum(doc_scores) / len(doc_scores)

    final_score = (norm_doc_score * avg_chunk_score + avg_chunk_score) / 2
    return final_score


def calc_chunk_scores(chunks: List[str], chunk_vectors: List[List[float]], query_vec: List[float]):
    """Calculates score of each chunk of a document

    Args:
        chunks: List of sentence chunks on the desired document
        chunk_vectors: List of vectors for each chunk in the document
        query_vec: vectorised query string

    Returns:
        Sorted list of chunks, with the highest first.
    """

    distances = cdist(
        [query_vec], chunk_vectors, "cosine")[0]
    scores = [1.0 - d for d in distances]
    chunk_scores = zip(chunks, scores)

    # Highest score first
    chunk_scores = sorted(chunk_scores, key=lambda x: x[1], reverse=True)

    return chunk_scores


def gen_preview_from_chunks(chunks: List[str], preview_length: int):
    """Generates a preview string for a document

    Args:
        chunks: List of strings, the chunks of the document
        preview_length: Number of characters in the preview

    Returns:
        Preview/metatext string
    """
    preview = ""
    reached_length = False
    for chunk in chunks:
        words = chunk.split()
        for word in words:
            if len(preview) + len(word) <= preview_length:
                preview += f" {word}"
            else:
                reached_length = True
                break

        if reached_length:
            break

    return preview


def process_results(search_results, query_vec, doc_class, preview_length: int,
                    max_results: int, min_score: float):
    """Calculates final scores for returned search results

    Args:
        search_results: Kiri SearchResults object.
        query_vec: Vectorised query string
        doc_class: Type of Document being processed
        preview_length: Number of characters in the preview string
    """
    max_score = -1.0

    for result in search_results.results:
        document = result.document
        # Only do processing for known document classes
        if issubclass(doc_class, ChunkedDocument):
            chunk_scores = calc_chunk_scores(document.chunks, document.chunk_vectors,
                                             query_vec)
            score = calc_doc_score(
                search_results.max_score, result.score, chunk_scores)
            result.score = score
            if score > max_score:
                max_score = score

            # Chunks ordered by highest score
            top_chunks = [cs[0] for cs in chunk_scores]
            preview = gen_preview_from_chunks(top_chunks, preview_length)
            result.preview = preview
        elif issubclass(doc_class, Document):
            # Treat entire content as a chunk
            chunk_scores = calc_chunk_scores([document.content], [document.vector],
                                             query_vec)
            score = calc_doc_score(
                search_results.max_score, result.score, chunk_scores)
            result.score = score
            if score > max_score:
                max_score = score

            preview = gen_preview_from_chunks(
                [document.content], preview_length)
            result.preview = preview

    # Order by decreasing score, up to max_results
    search_results.results = sorted(
        search_results.results, key=lambda r: r.score, reverse=True)[:max_results]

    # Remove documents with score < min_score
    search_results.results = [
        r for r in search_results.results if r.score >= min_score]

    search_results.max_score = max_score



import shortuuid
from typing import Dict, List, Tuple
from ..models import summarise, zero_shot, qa, emotion


class Document:
    """Base document class for extension as needed.

    Attributes:
        content: Text content of the document
        id: Unique ID for the document. Generated if not provided
        attributes: Dictionary of user-defined attributes
        vector: List of floats for doc vector representation

    Raises:
        TypeError: If content or id is not a string
        ValueError: If content or id is an empty string
    """

    def __init__(self, content: str, id: str = None,
                 attributes: Dict = None, vector: List[float] = None):
        if type(content) is not str:
            raise TypeError("content must be a string")

        if content == "":
            raise ValueError("content may not be the empty string ''")

        if id is None:
            id = shortuuid.uuid()

        if type(id) is not str:
            raise TypeError("id must be a string")

        if id == "":
            raise ValueError("id may not be the empty string ''")

        self.id = id
        self.content = content
        self.attributes = attributes
        self.vector = vector

    def summarise(self):
        # TODO: Handle long content
        return summarise(self.content)

    def classify(self, labels: List[str]):
        # TODO: Handle long content
        return zero_shot(self.content, labels)

    def qa(self, question, prev_qa: List[Tuple[str, str]] = []):
        # TODO: Handle long content
        return qa(question, self.content, prev_qa=prev_qa)

    def emotion(self):
        # TODO: Handle long content
        return emotion(self.content)

    def to_json(self, exclude_vectors=True):
        """Gets JSON form of the document.
        Returns:
            __dict__ of the document object
            """
        json_repr = vars(self)
        if exclude_vectors:
            del json_repr["vector"]
        return json_repr


class ChunkedDocument(Document):
    """Document subclass with support for finer-grained vector chunking.

    Attributes:
        args: Document superclass arguments
        chunking_level: 
        chunks: List of pre-chunked strings from the document
        chunk_vectors: List of vectors for each chunk of the document

    Raises:
        TypeError: If chunk_level is not an int, 
            chunks is not a list of strings, or 
            chunk_vectors is not a list of vectors
        ValueError: If chunk_level is < 1
    """

    def __init__(self, *args, chunking_level: int = 5, chunks: List[str] = None,
                 chunk_vectors: List[List[float]] = None, **kwargs):
        super(ChunkedDocument, self).__init__(*args, **kwargs)
        if type(chunking_level) != int:
            raise TypeError("chunking_level must be an int")

        if chunking_level < 1:
            raise ValueError("chunking_level must be >= 1")

        if chunks and type(chunks) is not list:
            raise TypeError("chunks must be a list of strings")

        if chunk_vectors and type(chunk_vectors) is not list:
            raise TypeError("chunk_vectors must be a list of vectors")

        self.chunking_level = chunking_level
        self.chunks = chunks
        self.chunk_vectors = chunk_vectors

    def to_json(self, exclude_vectors=True):
        """Gets JSON form of the document
        Returns:
            __dict__ attr of the document object
            """
        json_repr = vars(self)
        if exclude_vectors:
            del json_repr["vector"]
            del json_repr["chunk_vectors"]
        return json_repr


class ElasticDocument(Document):
    """Document with additional mapping required for ElasticSearch.

    Attributes:
        content: Text content of the document
        id: Unique ID for the document -- generated if not provided
        attributes: Dictionary of user-defined attributes
        vector: List of floats for doc vector representation
    """

    @staticmethod
    def elastic_mappings(dims=768):
        """Gets mappings for Elastic index

        Args:
            dims: Dimensions of document vector

        Returns:
            Dictionary of Elastic metadata 
        """
        # TODO: determine dims automatically
        mappings = {
            "properties": {
                "vector": {
                    "type": "dense_vector",
                    "dims": dims
                },
                "content": {
                    "type": "text"
                },
                "attributes": {
                    "type": "object"
                }
            }
        }

        return mappings

    def to_elastic(self):
        """Turns Document object into JSON form for Elastic

        Returns:
            __dict__ attribute of ElasticDocument object
        """
        return vars(self)

    @classmethod
    def from_elastic(cls, *args, **kwargs):
        obj = cls.__new__(cls)

        super(ElasticDocument, obj).__init__(*args, **kwargs)
        return obj


class ElasticChunkedDocument(ChunkedDocument, ElasticDocument):
    """Elastic-ready document with chunked vectorisation

    Attributes:
        args: Document & ChunkedDocument superclass arguments
    """

    def __init__(self, *args, **kwargs):
        super(ElasticChunkedDocument, self).__init__(*args, **kwargs)

    @staticmethod
    def elastic_mappings(dims=768):
        """Get mappings for elastic index

        Args:
            dims: Dimensions of document's vectors

        Returns:
            Dictionary of Elastic metadata 
        """
        # TODO: determine dims automatically
        mappings = {
            "properties": {
                "vector": {
                    "type": "dense_vector",
                    "dims": dims
                },
                "content": {
                    "type": "text"
                },
                "attributes": {
                    "type": "object"
                },
                "chunk_vectors": {
                    "type": "nested",
                    "properties": {
                        "vector": {
                            "type": "dense_vector",
                            "dims": dims
                        }
                    }
                }
            }
        }

        return mappings

    def to_elastic(self):
        """Turns Document object into JSON form for Elastic

        Returns:
            __dict__ attribute of ElasticDocument object
        """
        json_repr = vars(self)
        json_repr["chunk_vectors"] = [{"vector": v}
                                      for v in json_repr["chunk_vectors"]]
        return json_repr

    @classmethod
    def from_elastic(cls, *args, **kwargs):
        obj = cls.__new__(cls)

        kwargs["chunk_vectors"] = [v.get("vector")
                                   for v in kwargs["chunk_vectors"]]

        super(ElasticChunkedDocument, obj).__init__(*args, **kwargs)
        return obj



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kiri",
    version="0.2.0",
    author="Kiri OÜ",
    author_email="hello@kiri.ai",
    description="Kiri Natural Language Engine",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kiri-ai/kiri",
    packages=setuptools.find_packages(),
    install_requires=[
        "sentence_transformers",
        "nltk",
        "scipy",
        "shortuuid",
        "elasticsearch"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
